

/**
@author reconstruct by d
@since 2006/01/18
@note 
This is the original introductory comment:
-----------------------------------------

   A C-program for MT19937-64 (2004/9/29 version).
   Coded by Takuji Nishimura and Makoto Matsumoto.

   This is a 64-bit version of Mersenne Twister pseudorandom number
   generator.

   Before using, initialize the state by using init_genrand64(seed)  
   or init_by_array64(init_key, key_length).

   Copyright (C) 2004, Makoto Matsumoto and Takuji Nishimura,
   All rights reserved.                          

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   References:
   T. Nishimura, ``Tables of 64-bit Mersenne Twisters''
     ACM Transactions on Modeling and 
     Computer Simulation 10. (2000) 348--357.
   M. Matsumoto and T. Nishimura,
     ``Mersenne Twister: a 623-dimensionally equidistributed
       uniform pseudorandom number generator''
     ACM Transactions on Modeling and 
     Computer Simulation 8. (Jan. 1998) 3--30.

   Any feedback is very welcome.
   http://www.math.hiroshima-u.ac.jp/~m-mat/MT/emt.html
   email: m-mat @ math.sci.hiroshima-u.ac.jp (remove spaces)
@code

///Test Code
int main(void)
{
    int i;
    uint64 init[4]={0x12345, 0x23456	, 0x34567, 0x45678}, length=4;
    DKC_MT_RAND64 ctx;
		dkcMTRand64InitByArray(&ctx,init, length);
    printf("1000 outputs of dkcMTRand64Get()\n");
    for (i=0; i<1000; i++) {
      printf("%20llu ", dkcMTRand64Get(&ctx));
      if (i%5==4) printf("\n");
    }
    printf("\n1000 outputs of dkcMTRand64_real2()\n");
    for (i=0; i<1000; i++) {
      printf("%10.8f ", dkcMTRand64_real2(&ctx));
      if (i%5==4) printf("\n");
    }
    return 0;
}

@endcode
*/
#define DKUTIL_C_MT_RAND64_C
#include "dkcMTRand64.h"


#define NN dkcdMT_RAND64_NN
#define MM 156
#define MATRIX_A dkcmUINT64DEFINE(0xB5026F5AA96619E9)
#define UM dkcmUINT64DEFINE(0xFFFFFFFF80000000) /* Most significant 33 bits */
#define LM dkcmUINT64DEFINE(0x7FFFFFFF) /* Least significant 31 bits */

DKC_MT_RAND64 *WINAPI dkcAllocMTRand64(uint64 seed)
{
	const uint32 alignMask = 15;	/* = 2^n - 1 */
	uint8 *tp = dkcAllocate_INL(sizeof(DKC_MT_RAND64) + alignMask);
	DKC_MT_RAND64 *p;
	if(NULL==tp) return NULL;
	//mymt.c16byte align technic for SSE
	p = (DKC_MT_RAND64 *)(((int)tp + alignMask) & ~alignMask);
	p->basep = (DKC_MT_RAND64 *)tp;
	dkcMTRand64Init(p,seed);
	return p;
}

int WINAPI dkcFreeMTRand64(DKC_MT_RAND64 **p)
{
	return dkcFree_INL((void **)(&((*p)->basep)));
}

void WINAPI dkcMTRand64Init(DKC_MT_RAND64 *p,uint64 seed){
	int mti = p->mti;
	uint64 *mt = p->mt;
	p->mt[0] = seed;
  for (mti=1; mti<NN; mti++)
	{
		mt[mti] = (dkcmUINT64DEFINE(6364136223846793005) * (mt[mti-1] ^ (mt[mti-1] >> 62)) + mti);
	}
	p->mti = mti;
}

void WINAPI	dkcMTRand64InitByArray(DKC_MT_RAND64 *p,uint64 *init_key,uint64 key_length)
{
    uint64 i, j, k;
		uint64 *mt = p->mt;
    dkcMTRand64Init(p,dkcmUINT64DEFINE(19650218));
    i=1; j=0;
    k = (NN>key_length ? NN : key_length);
    for (; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 62)) * dkcmUINT64DEFINE(3935559000370003845)))
          + init_key[j] + j; /* non linear */
        i++; j++;
        if (i>=NN) { mt[0] = mt[NN-1]; i=1; }
        if (j>=key_length) j=0;
    }
    for (k=NN-1; k; k--) {
        mt[i] = (mt[i] ^ ((mt[i-1] ^ (mt[i-1] >> 62)) * dkcmUINT64DEFINE(2862933555777941757)))
          - i; /* non linear */
        i++;
        if (i>=NN) { mt[0] = mt[NN-1]; i=1; }
    }

    mt[0] = dkcmUINT64DEFINE(1) << 63; /* MSB is 1; assuring non-zero initial array */ 
}

DKC_INLINE void mt_rand64_next_c(DKC_MT_RAND64 *p){
	int i;
	int mti = p->mti;
	uint64 *mt = p->mt;
	uint64 x;
	/* if init_genrand64() has not been called, */
	/* a default initial seed is used     */
	/*if (mti == NN+1) 
			dkcMTRand64Init(p,dkcmUINT64DEFINE(5489)); 
	*/

	static uint64 mag01[2]={dkcmUINT64DEFINE(0), MATRIX_A};
	for (i=0;i<NN-MM;i++) {
			x = (mt[i]&UM)|(mt[i+1]&LM);
			mt[i] = mt[i+MM] ^ (x>>1) ^ mag01[(int)(x&dkcmUINT64DEFINE(1))];
	}
	for (;i<NN-1;i++) {
			x = (mt[i]&UM)|(mt[i+1]&LM);
			mt[i] = mt[i+(MM-NN)] ^ (x>>1) ^ mag01[(int)(x&dkcmUINT64DEFINE(1))];
	}
	x = (mt[NN-1]&UM)|(mt[0]&LM);
	mt[NN-1] = mt[MM-1] ^ (x>>1) ^ mag01[(int)(x&dkcmUINT64DEFINE(1))];

	mti = 0;

	//update
	p->mti = mti;
}

/* generates a random number on [0, 2^64-1]-interval */
uint64 WINAPI dkcMTRand64Get(DKC_MT_RAND64 *p)
{

    uint64 x;
	
    if (p->mti >= NN) { /* generate NN words at one time */
			mt_rand64_next_c(p);
    }
    x = p->mt[(p->mti)++];
	
    x ^= (x >> 29) & dkcmUINT64DEFINE(0x5555555555555555);
    x ^= (x << 17) & dkcmUINT64DEFINE(0x71D67FFFEDA60000);
    x ^= (x << 37) & dkcmUINT64DEFINE(0xFFF7EEE000000000);
    x ^= (x >> 43);
		
    return x;
}


//---------------------------------

///ϐmtiDKC_MT_RAND64_CACHE̊ǗɂB
DKC_INLINE void WINAPI dkcMTRand64CacheNext(DKC_MT_RAND64_CACHE *p){
	
	int i;
	mt_rand64_next_c(p->ptr);
	for(i=0;i<NN;i++){
		p->cache[i] = dkcMTRand64Get(p->ptr);
	}
	p->ptr->mti = 0;
}

